/*
** identd.c                   A TCP/IP link identification protocol server
**
** Version: 1.0.0
**
** This program is in the public domain and may be used freely by anyone
** who wants to.
**
** Last update: 20 Sep 1993
**
** Please send bug fixes/bug reports to:
**                  Pauline Middelink (middelin@calvin.iaf.nl)
*/

/*
   modified: 17 Dec 1996 by widi
      (widiyanto Andiman, wd@pegasus.dvz.fh-aachen.de
      or widiyanto-andiman@dialup.fh-aachen.de )

   how does it work ???

   If the -f argument is given, it immediatly sends an user identification
   as a parameter given by the argument -u (see belaow),
   or DEFAULT_USERID if omited.
   (WARNING: DO NOT use both argument -f and -u for ethical purposes)

   In a normal case:
   it looks up current connections in /proc/net/tcp, see manual.
   if found then sends response.
   if no such connection then looks up in /proc/net/ip_masquerade.
   if not found, sends userid as parameter given by argument -u
   or error if omited,
   otherwise, it requests an identification to the source host
   indicated by this maquerade entry.
   Any response from the source host will be send by repleacing
   it's port pair to requester.

   new arguments has been added:
   -f		: force sending response w/ specified user (see below)
        	  or default user if no user is given
   -u [user]	: specifiy user as response in case of:
        	  i) arguments -f is used
                 ii) any error on querying to client (ip masquerade case)

                 if 'user' is omited, default user is 'any'
                 or whatever you define as DEFAULT_USERID, see below
*/

#define VERSION "[identd, version 1.0.0]"

#define DEFAULT_USERID "any"

#include <stdio.h>
#include <netdb.h>
#include <pwd.h>
#include <syslog.h>
#include <sys/types.h>
#include <sys/param.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>

int debug_flag = 0;
int syslog_flag = 0;

/*
** Return the name of the connecting host, or the IP number as a string.
*/
const
char *gethost(addr)
struct in_addr *addr;
{
  struct hostent *hp;

  hp = gethostbyaddr((char *)addr, sizeof(struct in_addr), AF_INET);
  if (hp) return hp->h_name;
  else return inet_ntoa(*addr);
}

int getinfo(struct in_addr *faddr, int fport, struct in_addr *laddr, int lport)
{
   FILE      *fp;
   long      dummy;
   char      buf[132];
   struct in_addr   myladdr, myraddr;
   int      mylport, myrport;
   int      uid;

   /*
   ** Open the kernel memory device
   */
   if ((fp = fopen("/proc/net/tcp", "r"))==NULL) {
      if (syslog_flag)
         syslog(LOG_INFO, "fopen(/proc/net/tcp) failed");

                printf("%d, %d: ERROR: UNKNOWN-ERROR\r\n", lport, fport), \
      exit(1);
   }

   /* eat header */
   fgets(buf,sizeof(buf)-1,fp);
   while (fgets(buf, sizeof(buf)-1, fp)){
      if(sscanf(buf,"%d: %lX:%x %lX:%x %x %lX:%lX %x:%lX %lx %d",
         	&dummy, &myladdr, &mylport, &myraddr, &myrport,
                &dummy, &dummy, &dummy, &dummy, &dummy, &dummy,
                &uid)==12) {
         if (myladdr.s_addr == laddr->s_addr &&
            	mylport == lport && myrport == fport){
            faddr->s_addr = myraddr.s_addr;
            return uid;
         }
      }
   }
   return -1;
}

/*
now looks up /proc/net/ip_masquerade
*/
int getinfo_masq(struct in_addr *caddr, int *cport, int fport, int lport)
{
   FILE *f;
   long dummy;
   char buf[256];
   struct in_addr clientaddr;
   int masqport;

   if((f = fopen("/proc/net/ip_masquerade", "r")) == NULL){
      if (syslog_flag)
         syslog(LOG_INFO, "fopen(/proc/net/ip_masquerade) failed");
      return -1;
   }
   if(syslog_flag && debug_flag)
      syslog(LOG_INFO, "scans fport %x masq port %x", fport, lport);

   fgets(buf, 255, f);
   while(fgets(buf, 255, f)){
      if(sscanf(buf, "%s %lX:%x %lX:%x %x",
         	buf, &clientaddr, cport, &dummy, &dummy, &masqport) == 6){
         if(lport == masqport) {
            caddr->s_addr = ntohl(clientaddr.s_addr);
            return 0;
         }
      }
   }
   return -1;
}

int query_client_ident(struct in_addr *caddr, int cport, int fport,
   char *response)
{
   FILE *f;
   int sock;
   register int c;
   struct sockaddr_in sn;
   struct servent *sp;
   char buf[128];
   int ignore = 1;

   if(syslog_flag && debug_flag)
      syslog(LOG_INFO, "connecting to client, asking port %d", cport);

   sp = getservbyname("auth", "tcp");
   if(!sp){
      if(syslog_flag)
         syslog(LOG_INFO, "unknown auth service");
      return -1;
   }

   if((sock = socket(AF_INET, SOCK_STREAM, 0)) < 0){
      if(syslog_flag)
         syslog(LOG_INFO, "error opening socket");
      return -1;
   }

   sn.sin_family = AF_INET;
   memcpy(&sn.sin_addr, caddr, sizeof(struct in_addr));
   sn.sin_port = sp->s_port; // ident port
   memset(sn.sin_zero, 0, sizeof(sn.sin_zero));
   if(connect(sock, (struct sockaddr *) &sn, sizeof(sn)) < 0){
      if(syslog_flag)
         syslog(LOG_INFO, "error connect to client");
      close(sock);
      return -1;
   }
   sprintf(buf, "%d, %d\r\n", cport, fport);
   write(sock, buf, strlen(buf));

   /* now read the response */
   if((f = fdopen(sock, "r")) == NULL){
      if(syslog_flag)
         syslog(LOG_INFO, "error fdopen");
      close(sock);
      return -1;
   }
   /* reads as is beginning from 1st ':', also ':' is also included */
   while((c = getc(f)) != EOF){
      if((char) c == ':') ignore = 0;
      if(!ignore) *response++ = (char) c;
   }
   *response = '\0';
   fclose(f);
   close(sock);
   return 0;
}

int main(int argc, char *argv[]) {
   int      i, len;
   int      uid;
   struct sockaddr_in sin;
   struct in_addr   laddr, faddr;
   int      lport, fport;
   struct passwd   *pwp;
   char *user = NULL;
   int userflag = 0;
   int forceuserflag = 0;

   /*
    * Parse the command line arguments
    */
   for (i = 1; i < argc && argv[i][0] == '-'; i++)
      while (*(++argv[i]))
         switch (*argv[i]) {
         case 'l': syslog_flag++;
              break;
         case 'V': puts(VERSION);
              exit(0);
         case 'd': debug_flag++;
              break;
         case 'f':
            forceuserflag = 1;
            break;
         case 'u' :
            userflag = 1;
            user = argv[i+1];
            if(user) i++;
            break;
         }

   /*
    * Get foreign internet address
    */
   len = sizeof(sin);
   if (getpeername(0,(struct sockaddr *)&sin,&len) == -1) {
      /*
       * A user has tried to start us from the command line or
       * the network link died, in which case this message won't
       * reach to other end anyway, so lets give the poor user some
       * errors.
       */
      perror("identd: getpeername()");
      exit(1);
   }

   faddr = sin.sin_addr;

   /*
    * Get local internet address
    */
   len = sizeof(sin);
   if (getsockname(0,(struct sockaddr *)&sin,&len) == -1) {
      /*
       * We can just die here, because if this fails then the
       * network has died and we haven't got anyone to return
       * errors to.
       */
      exit(1);
   }
   laddr = sin.sin_addr;

   /*
    * Open the connection to the Syslog daemon if requested
    */
   if (syslog_flag) {
#ifdef LOG_DAEMON
      openlog("identd", LOG_PID, LOG_DAEMON);
#else
      openlog("identd", LOG_PID);
#endif
      /*
      syslog(LOG_NOTICE, "Connection from %s", gethost(&faddr));
      */
   }

   /*
    * Get the local/foreign port pair from the luser
    */
   lport = fport = 0;
   if (scanf("%d , %d", &lport, &fport) != 2 ||
       lport < 1 || lport > 65535 ||
        fport < 1 || fport > 65535) {
      if (syslog_flag)
         syslog(LOG_INFO, "scanf: invalid-port(s): %d, %d", lport, fport);

      printf("%d, %d: ERROR: INVALID-PORT\r\n", lport, fport);
      exit(0);
   }

   /* force with specified user ?? */
   /* you should not use this way !! */
   if(forceuserflag){
      if(user == NULL) user = DEFAULT_USERID;
      if(syslog_flag)
         syslog(LOG_INFO, "Force response as: %d, %d: %s", lport, fport, user);
      printf("%d, %d: USERID: UNIX: %s\r\n", lport, fport, user);
      exit(0);
   }

      /*
    * Next - get the specific TCP connection and return the uid
    */
   uid = getinfo(&faddr, fport, &laddr, lport);

   if(syslog_flag){
      syslog(LOG_NOTICE, "Connection from %s", gethost(&faddr));
      syslog(LOG_INFO, "fport: %d lport: %d", fport, lport);
   }

   if (uid<0) {
      /* trying w/ ip-masquerade entries */
      struct in_addr caddr;
      int cport;

      /* look at masquerade entries */
      if(getinfo_masq(&caddr, &cport, fport, lport) == 0){
         /* we have found it, we got also client addr and port,
            so request ident to client host */
         char response[128];
         char *sz;

         if(syslog_flag)
            syslog(LOG_NOTICE, "Reconnects to client %s", gethost(&caddr));

         if(query_client_ident(&caddr, cport, fport, response) < 0){
            /* error */
            /* check if 'u' paramter is given */
            if(userflag){
               if(user == NULL) user = DEFAULT_USERID;
               printf("%d, %d: USERID: UNIX: %s\r\n", lport, fport, user);
               exit(0);
            }
         /* otherwise */
            printf("%d, %d: ERROR: UNKNOWN-ERROR\r\n", lport, fport);
            exit(1);
         }
         for(sz = response; *sz && *sz >= ' '; sz++);
         *sz = '\0';
         if (syslog_flag && debug_flag)
            syslog(LOG_INFO, "Successful client lookup: %d, %d %s",
                   lport, fport, response);
         /* !!! lport is correct here !!!! */
         printf("%d, %d %s\r\n", lport, fport, response);
         exit(0);
      }
      else{ /* not found ! */

         if (syslog_flag && debug_flag)
            syslog(LOG_INFO, "Returned: %d, %d: NO-USER", lport, fport);

         printf("%d, %d: ERROR: NO-USER\r\n", lport, fport);
         exit(0);
      }
   }

   /*
    * Then we should try to get the username
    */
   pwp = getpwuid(uid);
   if (!pwp) {
      if (syslog_flag)
         syslog(LOG_WARNING, "getpwuid() could not map uid (%d) to name", uid);

      printf("%d, %d: ERROR: UNKNOWN-ERROR\r\n", lport, fport);
      exit(1);
   }

   /*
    * Hey! We finally made it!!!
    */
   if (syslog_flag && debug_flag)
      syslog(LOG_INFO, "Successful lookup: %d, %d: %s", lport, fport, pwp->pw_name);

   printf("%d, %d: USERID: UNIX: %s\r\n", lport, fport, pwp->pw_name);
   exit(0);
}
